/*  XMPS - X MPEG Player System
 *  Copyright (C) 1999 Damien Chavarria
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public Licensse as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/**
 *
 * win32_codec.c
 *
 * WIN32 DLL Codec.
 * @Author Chavarria Damien.
 *         Eugene Smith.
 * 	   Copyright 1999-2000
 */

#include "win32_codec.h"

typedef struct {

  xmps_system_plugin_t   *our_system;
  unsigned int            our_stream_id;
  xmps_video_info_t      *video_info;
  HIC                     hic;
  BITMAPINFOHEADER        bih;
  struct my_BITMAPINFO    output_bih;
  HRESULT                 h;
  void                   *our_out_buffer;
  char                   *codec_path;
  unsigned int           *next_size;

} win32_data_t;

/*
 * get_video_codec_info : MANDATORY
 *
 * Used by scanplugins to get a pointer to the codec.
 *
 */

xmps_video_decoder_plugin_t *get_video_decoder_info(void)
{
  xmps_video_decoder_plugin_t *video_decoder;

  video_decoder = (xmps_video_decoder_plugin_t *) malloc(sizeof(xmps_video_decoder_plugin_t));

  video_decoder->name = "WIN32";
  video_decoder->data = malloc(sizeof(win32_data_t));

  video_decoder->open       = win32_open;
  video_decoder->set        = win32_set;
  video_decoder->get        = win32_get;
  video_decoder->decompress = win32_decompress;
  video_decoder->close      = win32_close;

  
  ((win32_data_t *) video_decoder->data)->video_info = (xmps_video_info_t *) malloc(sizeof(xmps_video_info_t));
  ((win32_data_t *) video_decoder->data)->our_system = NULL;
  ((win32_data_t *) video_decoder->data)->codec_path = "/usr/lib/win32";

  return video_decoder;
}

unsigned int win32_open(xmps_video_decoder_plugin_t *video_decoder)
{
  if(video_decoder == NULL) {
    return 0;
  }

  return 1;
}

void* win32_get(xmps_video_decoder_plugin_t *video_decoder, unsigned int flag, void *arg)
{
  win32_data_t *data;

  if(video_decoder == NULL) {
    return 0;
  }

  data = (win32_data_t *) video_decoder->data;
 
  if(data == NULL) {
    return 0;
  }

  switch(flag)
    {
    case XMPS_FLAG_XML_CONFIG:
      {
	GList *list = NULL;
  
	list = g_list_prepend(list, (void *) xmps_xml_config_entry_string_new("codec_path", data->codec_path, "dir", "Path to win32 DLLs:"));

	return list;
      }
      break;

    case XMPS_FLAG_VIDEO_INFO:
      {
	
	if(data->our_system == NULL)
	  return NULL;
	
	return (void *) data->video_info;
      }
      break;
    case XMPS_FLAG_VIDEO_FORMAT_LIST:
      {
	GList             *formats = NULL;

	if(data->our_system == NULL)
	  return NULL;

	data->video_info->format.type = XMPS_VIDEO_FORMAT_RGB24;
	data->video_info->format.bpp  = 24;
	
	formats = g_list_append(formats, &data->video_info->format);
  
	return (void *) formats;
      }
      break;
    default:
      break;
    }
  
  return NULL;
}

void win32_parse_xml(xmps_video_decoder_plugin_t *video_decoder, GList * xml)
{
  win32_data_t *data;
  GList        *node;
  char         *xml_data;

  if(video_decoder == NULL) {

    XMPS_DEBUG("wrong params!");
    return;
  }

  data = (win32_data_t *) video_decoder->data;

  if(data == NULL) {

    XMPS_DEBUG("error!");
    return;
  }

  for(node = xml; node != NULL; node = g_list_next(node)) {

    xml_data = (char *) node->data;

    if(xml_data == NULL) {
      break;
    }
    
    if(strstr(xml_data, "codec_path") != NULL) {

      XMPS_DEBUG("updating codec path");
      
      data->codec_path = xmps_xml_get_value(xml_data, "value");
    }
  }

  XMPS_DEBUG("codec_path = %s", data->codec_path);
}



unsigned int win32_set(xmps_video_decoder_plugin_t *video_decoder, unsigned int flag, void *arg)
{
  win32_data_t *data;

  if(video_decoder == NULL) {

    XMPS_DEBUG("problem!");
    return 0;
  }

  data = (win32_data_t *) video_decoder->data;
 
  if(data == NULL) {
    
    XMPS_DEBUG("problem!");
    return 0;
  }

  switch(flag)
    {

    case XMPS_FLAG_SAVE_CONFIG:
      {
	xmps_config_file_t *cfgfile;
	
	cfgfile = (xmps_config_file_t *) arg;

	xmps_config_write_string(cfgfile, 
				 "win32_decoder", 
				 "codec_path", 
				 data->codec_path);

	xmps_config_write_default_file(cfgfile);
	
	return 1;
      }
      break;
      
    case XMPS_FLAG_XML_CONFIG:

      XMPS_DEBUG("updating config");

      win32_parse_xml(video_decoder, (GList *) arg);

      return 1;
      break;

    case XMPS_FLAG_INPUT:
      {
	xmps_data_t *data_stream;

	data_stream = (xmps_data_t *) arg;

	if(data_stream == NULL) {
	  return 0;
	}

	XMPS_DEBUG("trying to open data stream!");
	
	if(data->our_system != NULL) {

	  XMPS_DEBUG("our system is not NULL");
	  data->our_system = NULL;
	}
	
	if(data_stream->plugin_type == XMPS_PLUGIN_SYSTEM) {
	  data->our_system = data_stream->plugin;
	}

	if(data->our_system == NULL) {
	  return 0;
	}

	if(data_stream->type == XMPS_DATA_VIDEO_COMP && data_stream->video_compression != NULL) {
	  
	  XMPS_DEBUG("open : stream is video");

	  if( strstr("IV50", data_stream->video_compression) != NULL ||
	      strstr("IV40", data_stream->video_compression) != NULL ||
	      strstr("IV32", data_stream->video_compression) != NULL ||
	      strstr("DIV3", data_stream->video_compression) != NULL || 
	      strstr("DIV4", data_stream->video_compression) != NULL || 
	      strstr("div3", data_stream->video_compression) != NULL || 
	      strstr("div4", data_stream->video_compression) != NULL ||
	      strstr("mp43", data_stream->video_compression) != NULL || 
	      strstr("mp41", data_stream->video_compression) != NULL ) 
	    {
	      
	      /*
	       * setting codec path
	       *
	       */
	      
	      XMPS_DEBUG("setting codec path to %s", data->codec_path);

	      SetCodecPath(data->codec_path);
	      
	      XMPS_DEBUG("found 1 video stream with compression %s", 
			 data_stream->video_compression);
	      
	      if(data_stream->user_data == NULL) {
		
		data->bih.biSize        = sizeof(BITMAPINFOHEADER);
		data->bih.biWidth       = data_stream->video_info.width;
		data->bih.biHeight      = data_stream->video_info.height;
		data->bih.biPlanes      = 1;
		data->bih.biBitCount    = 24;
		data->bih.biSizeImage   = abs(data->bih.biWidth * data->bih.biHeight * data->bih.biBitCount / 8);
		data->bih.biCompression = ((unsigned char) *data_stream->video_compression) | 
		  ((unsigned char) *(data_stream->video_compression + 1)) <<  8 |
		  ((unsigned char) *(data_stream->video_compression + 2)) << 16 |
		  ((unsigned char) *(data_stream->video_compression + 3)) << 24;
		data->bih.biClrUsed      = 0;
		data->bih.biClrImportant = 0;
	      }
	      else {
		
		memcpy(&data->bih, data_stream->user_data, sizeof(BITMAPINFOHEADER));
	      }

	      memset(&data->output_bih, 0, sizeof(struct my_BITMAPINFO));
	      data->output_bih.header.biSize = sizeof(BITMAPINFOHEADER);
	      
	      XMPS_DEBUG("open : building BITMAPINFOHEADER");
	
	      data->hic = ICOpen( 0x63646976, 
				  data->bih.biCompression, 
				  ICMODE_FASTDECOMPRESS);
	
	      if(data->hic == 0) {

		XMPS_DEBUG("ICOpen failed : wrong parameters?");
		return 0;
	      }
	      
	      data->video_info->width  = data->bih.biWidth;
	      data->video_info->height = data->bih.biHeight;

	      data->video_info->format.type = XMPS_VIDEO_FORMAT_RGB24;
	      data->video_info->format.bpp  = 24;

	      /*
	       * now init decompression
	       *
	       */

	      XMPS_DEBUG("setting output format");

	      if(data->hic == 0)
		return 0;

	      data->bih.biHeight = abs(data->bih.biHeight);

	      data->bih.biBitCount  = data->video_info->format.bpp;
	      data->bih.biSizeImage = abs(data->bih.biWidth * data->bih.biHeight * data->bih.biBitCount / 8);

	      data->h = ICDecompressGetFormat(data->hic, &data->bih, &data->output_bih);
	
	      if(data->h == NULL) {
		XMPS_DEBUG("ICDecompressGetFormat failed");
		ICClose(data->hic);
		return 0;
	      }
	
	      data->h = ICDecompressQuery(data->hic, &data->bih, &data->output_bih);
	
	      if(data->h == NULL) {
		
		XMPS_DEBUG("ICDecompressQuery failed");
		

		ICClose(data->hic);
		return 0;
	      }
	      
	      XMPS_DEBUG("changing bit depth to %d", 
			 data->output_bih.header.biBitCount);

	      data->output_bih.header.biBitCount = 24;

	      data->h = ICDecompressBegin(data->hic, &data->bih, &data->output_bih);
	      
	      if(data->h == NULL) {
		
		XMPS_DEBUG("ICDecompressDegin failed");
		
		data->h = ICDecompressEnd(data->hic);
		ICClose(data->hic);
		return 0;
	      }           

	      XMPS_DEBUG("allocating buffer of %dx%d, %dbpp", 
			 data->video_info->width, data->video_info->height, data->video_info->format.bpp);
	      
	      if(data->our_out_buffer == NULL) {
		data->our_out_buffer = malloc(data->video_info->width*data->video_info->height*data->video_info->format.bpp/8); 
	      }
	      else {
		data->our_out_buffer = realloc(data->our_out_buffer, data->video_info->width*data->video_info->height*data->video_info->format.bpp/8); 
	      }

	      return 1;
	    }
	  else {
	    
	    data->our_system = NULL;
	    return 0;
	  }
	}
	else {
	  return 0;
	}
	
      }
      break;

    case XMPS_FLAG_DROPFRAME:
      
      if(data->next_size == NULL) {
	
	data->next_size = malloc(sizeof(unsigned int));
	*data->next_size = 0;
      }
      
      data->bih.biSizeImage = *data->next_size;

      data->h = ICDecompress(data->hic,
			     ICDECOMPRESS_HURRYUP,
			     &data->bih,
			     arg,
			     (BITMAPINFOHEADER *) &data->output_bih, 
			     data->our_out_buffer);

      return 1;
      break;

    default:
      break;
    }

  return 0;
}

unsigned int win32_decompress(xmps_video_decoder_plugin_t *video_decoder, void *in_buffer, void *out_buffer, unsigned int size)
{
  int i, line_length;
  win32_data_t *data;

  if(video_decoder == NULL) {

    XMPS_DEBUG("null_video_decoder!");
    return 0;
  }

  data = (win32_data_t *) video_decoder->data;
 
  if(data == NULL) {

    XMPS_DEBUG("null data");
    return 0;
  }

  if(data->hic == 0) {
    XMPS_DEBUG("decompress : codec not initialized");
    return 0;
  }

  if(data->our_system == NULL) {

    XMPS_DEBUG("decompress : system not ready");
    return 0;
  }

  data->bih.biSizeImage = size;

  if(size == 0) {

    data->h = ICDecompress(data->hic, 
			   ICDECOMPRESS_NULLFRAME,
			   &data->bih, 
			   in_buffer, 
			   (BITMAPINFOHEADER*) &data->output_bih, 
			   data->our_out_buffer);
    
  }
  else {

    data->h = ICDecompress(data->hic, 
			   ICDECOMPRESS_NOTKEYFRAME, 
			   &data->bih, 
			   in_buffer, 
			   (BITMAPINFOHEADER*) &data->output_bih, 
			   data->our_out_buffer);
  }

  /*
   * now invert the image
   * 
   */
  
  line_length = data->bih.biWidth*data->bih.biBitCount/8;
    
  for(i=0; i < abs(data->bih.biHeight); i++) {
    
    memcpy(out_buffer + i*line_length, data->our_out_buffer + (ABS(data->bih.biHeight) - i - 1)*line_length, line_length);
  }

  data->next_size = data->our_system->get(data->our_system, XMPS_FLAG_VIDEO_BUFFER_SIZE, NULL);
					  
  return 1;
}

unsigned int win32_close(xmps_video_decoder_plugin_t *video_decoder)
{
  win32_data_t *data;
  
  if(video_decoder == NULL) {
    return 0;
  }
  
  data = (win32_data_t *) video_decoder->data;
 
  if(data == NULL) {
    return 0;
  }

  XMPS_DEBUG("close");
  
  data->our_system = NULL;
  
  data->h = ICDecompressEnd(data->hic);
  ICClose(data->hic);

  data->hic = 0;

  return 1;
}



